;;;; wh-engine/package.lisp
;;;; root package of wh-engine

(defpackage wh-engine
  (:nicknames whe)
  (:use common-lisp 3d-vectors 3d-matrices)
  (:import-from sb-ext
                weak-pointer weak-pointer-p make-weak-pointer weak-pointer-value)
  (:import-from objective-lisp O!)
  (:export
   ;; main.lisp
   +version+ ensure-version
   deref-pointer points-to ensure-live
   vxy1 vxy-trunc
   make-id fixed-id
   *world*
   add-scene remove-scene get-scene update-all-scenes
   initialize-actors-in
   register-test-scene
   run
   *delta-time*

   ;; serialization.lisp
   id-ref make-id-ref
   pointer
   deref-sus-pointer
   referize dereferize pointerize
   referize-setf pointerize-setf
   make-generic-load-form
   generate-load-forms
   dump-scene dump-actors
   load-resume-scene load-resume-actors

   ;; actor.lisp
   actor
   ; properties
   id name scene tags active-p parent children components destroyed-p
   location z-layer rotation scale matrix
   ; virtual properties
   tree-active-p
   world-matrix local-matrix
   world-location
   ; methods
   get-component add-component
   add-child remove-child
   apply-to-tree
   has-tag add-tag remove-tag
   parent-changed
   deactivate activate
   resume suspend
   update
   destroy
   transform-point world-point local-point
   transform-svector transform-vector
   translate-by rotate-by scale-by

   ;; component.lisp
   component
   ; properties
   actor active-p started-p
   ; virtual properties
   scene destroyed-p
   ; methods
   attach
   parent-changed
   deactivate activate
   resume suspend
   start update
   destroy

   ;; actor-macros.lisp
   new!
   define-new!-impl

   ;; scene.lisp
   scene
   ; properties
   id name actors destroyed-p
   ; methods
   add-actor remove-actor
   get-actor get-tagged-actors
   update
   destroy
   resume suspend

   ;; systems.lisp
   register-system install-systems
   ))
